/**
** $Id: streamdecoder.cpp,v 1.1.2.4 2012-12-01 16:40:56 fliegl Exp $
**/

#include <time.h>
#include <fcntl.h>
#include <stdio.h>
#include <stdarg.h>
#include <stdlib.h>
#include <string.h>
#include <getopt.h>
#include <unistd.h>

#include <neaacdec.h>
#include <stdint.h>
#include <curl/curl.h>
#include <mpg123.h>

#ifdef STANDALONE
#include <time.h>
#include <SDL/SDL.h>
#include <SDL/SDL_audio.h>
#include <bs1770_ctx.h>

SDL_AudioSpec desired;
SDL_AudioSpec obtained;

#define BUFSZ (512*1024)
char *pcm;

unsigned int writeptr=0;
unsigned int readptr=0;

bs1770_ctx_t *ctx;
time_t last_lufs=time(NULL),last_lra=time(NULL);

#define SDL

void sdl_callback(void *userdata, Uint8 *stream, int len) 
{
    unsigned int rd=readptr%BUFSZ;
    memset(stream, 0, len);
    if(writeptr-readptr < len) {
//        printf("nix %d %d\n", writeptr-readptr, len);
        return;
    }
//    printf("sdl_callback %d %d %d\n", readptr, rd, len);
    
    if(rd+len <= BUFSZ) {
        memcpy(stream, pcm+rd, len);
//        printf("->rd:%d len:%d\n",rd, len);
    } else {
        int todo=BUFSZ-rd;
        memcpy(stream, pcm+rd, todo);
//        printf("->1.rd:%d len:%d\n",rd, todo);
        memcpy(stream+todo, pcm, len-todo);
//        printf("->2.rd:%d len:%d\n",0, len-todo);
    }
    readptr+=len;
}

#if defined SDL
static void write_cb(const void *p, const void* buf, size_t len, int srate, int channels) 
{
    time_t t=time(NULL);

    bs1770_ctx_add_samples_i_i16(ctx, 0, (double)srate, channels, (bs1770_i16_t *)buf, len>>2);
        
    if(t-last_lufs>=10) {
        fprintf(stderr, "%.1f LUFS\n", bs1770_ctx_track_lufs_r128(ctx,0));
        last_lufs=t;
    }
    if(t-last_lra>=60) {
        fprintf(stderr, "%.1f LU(LRA)\n", bs1770_ctx_track_lra_default(ctx,0));
        last_lra=t;
    }
    SDL_LockAudio();
    unsigned int wr=writeptr%BUFSZ;
//    printf("write_cb ptr:%d len:%d BUFSZ:%d\n", writeptr, len, BUFSZ);
    if(wr+len <= BUFSZ) {
        memcpy(pcm+wr, buf, len);
//        printf("->wr:%d len:%d\n",wr, len);
    } else {
        int todo=BUFSZ-wr;
        memcpy(pcm+wr, buf, todo);
//        printf("->1.wr:%d len:%d\n",wr, todo);
        memcpy(pcm, (char *)buf+todo, len-todo);
//        printf("->2.wr:%d len:%d\n",0, len-todo);
    }
    writeptr+=len;
    SDL_UnlockAudio();
}
#else
static void write_cb(const void *p, const void* buf, size_t len, int srate, int channels) 
{
    write(1,buf,len);
}
 
#endif     
#endif

#include "fopen.h"
#include "streamdecoder.h"

#define MAX_CHANNELS 6 /* make this higher to support files with
                          more channels */


/* MicroSoft channel definitions */
#define SPEAKER_FRONT_LEFT             0x1
#define SPEAKER_FRONT_RIGHT            0x2
#define SPEAKER_FRONT_CENTER           0x4
#define SPEAKER_LOW_FREQUENCY          0x8
#define SPEAKER_BACK_LEFT              0x10
#define SPEAKER_BACK_RIGHT             0x20
#define SPEAKER_FRONT_LEFT_OF_CENTER   0x40
#define SPEAKER_FRONT_RIGHT_OF_CENTER  0x80
#define SPEAKER_BACK_CENTER            0x100
#define SPEAKER_SIDE_LEFT              0x200
#define SPEAKER_SIDE_RIGHT             0x400
#define SPEAKER_TOP_CENTER             0x800
#define SPEAKER_TOP_FRONT_LEFT         0x1000
#define SPEAKER_TOP_FRONT_CENTER       0x2000
#define SPEAKER_TOP_FRONT_RIGHT        0x4000
#define SPEAKER_TOP_BACK_LEFT          0x8000
#define SPEAKER_TOP_BACK_CENTER        0x10000
#define SPEAKER_TOP_BACK_RIGHT         0x20000
#define SPEAKER_RESERVED               0x80000000

static int init_done = 0;

static void(*stream_cb)(const void*, const void*, size_t, int, int)=NULL;

void register_stream_callback( void(*stream_callback)(const void *, const void*, size_t, int, int))
{
    stream_cb=stream_callback;
}

static long aacChannelConfig2wavexChannelMask(NeAACDecFrameInfo *hInfo)
{
    if (hInfo->channels == 6 && hInfo->num_lfe_channels)
    {
        return SPEAKER_FRONT_LEFT + SPEAKER_FRONT_RIGHT +
            SPEAKER_FRONT_CENTER + SPEAKER_LOW_FREQUENCY +
            SPEAKER_BACK_LEFT + SPEAKER_BACK_RIGHT;
    } else {
        return 0;
    }
}

static void print_channel_info(NeAACDecFrameInfo *frameInfo)
{
    /* print some channel info */
    int i;
    long channelMask = aacChannelConfig2wavexChannelMask(frameInfo);
    fprintf(stderr, "Format AAC %luHz ", frameInfo->samplerate);
    if (frameInfo->num_lfe_channels > 0)
    {
        fprintf(stderr, "%2d.%dch\n", frameInfo->channels-frameInfo->num_lfe_channels, frameInfo->num_lfe_channels);
    } else {
        fprintf(stderr, "%2dch\n", frameInfo->channels);
    }
}

static int decodeAACstream(decoder_context_t *ctx, URL_FILE *handle)
{
    unsigned long samplerate;
    unsigned char channels;
    void *sample_buffer;
    unsigned char *aac_buffer;

    NeAACDecHandle hDecoder;
    NeAACDecFrameInfo frameInfo;
    NeAACDecConfigurationPtr config;

    int bread;
    int bdecoded;

    int first_time = 1;


    if (!(aac_buffer = (unsigned char*)malloc(FAAD_MIN_STREAMSIZE*MAX_CHANNELS)))
    {
        fprintf(stderr, "Memory allocation error\n");
        return 0;
    }
    memset(aac_buffer, 0, FAAD_MIN_STREAMSIZE*MAX_CHANNELS);

    bread = url_fread(aac_buffer, 1, FAAD_MIN_STREAMSIZE*MAX_CHANNELS, handle);


    hDecoder = NeAACDecOpen();
#if 0
    config = NeAACDecGetCurrentConfiguration(hDecoder);
//  config->defSampleRate = def_srate;
    config->defObjectType = HE_AAC;
//    config->outputFormat = outputFormat;
//    config->downMatrix = downMatrix;
//    config->useOldADTSFormat = old_format;
//    config->dontUpSampleImplicitSBR = 1;
    NeAACDecSetConfiguration(hDecoder, config);
#endif
    if ((bdecoded = NeAACDecInit(hDecoder, aac_buffer, bread, &samplerate, &channels)) < 0)
    {
        /* If some error initializing occured, skip the file */
        fprintf(stderr, "Error initializing decoder library.\n");
        if (aac_buffer)
            free(aac_buffer);
        NeAACDecClose(hDecoder);
        url_fclose(handle);
        return 1;
    }

    do
    {
        sample_buffer = NeAACDecDecode(hDecoder, &frameInfo, aac_buffer, bread);
        if (frameInfo.error > 0)
        {
            fprintf(stderr, "Error: %s\n",
                NeAACDecGetErrorMessage(frameInfo.error));
        }
        /* open the sound file now that the number of channels are known */
        if (first_time && !frameInfo.error)
        {
            /* print some channel info */
            print_channel_info(&frameInfo);
#if defined STANDALONE and defined SDL
	    desired.freq = frameInfo.samplerate;
    	    desired.format = AUDIO_S16;
    	    desired.channels = frameInfo.channels;
            desired.samples = 960;        /* A good value for games */
            desired.callback = sdl_callback;
            desired.userdata = NULL;
            if ( SDL_OpenAudio(&desired, &obtained) < 0 ){
                fprintf(stderr, "Couldn't open audio: %s\n", SDL_GetError());
                exit(-1);
            }
            SDL_PauseAudio(0);
#endif
            first_time = 0;
        }
//        printf("bread: %d, consumed: %lu %p\n", bread, frameInfo.bytesconsumed, sample_buffer);
        stream_cb(ctx->p, sample_buffer, frameInfo.samples<<1, frameInfo.samplerate, frameInfo.channels );
        int newpos=bread-frameInfo.bytesconsumed;
        
        if(newpos) {
            memmove(aac_buffer, aac_buffer+frameInfo.bytesconsumed, newpos);
            bread=newpos;  
        }
        int tmpbread = url_fread(aac_buffer+bread, 1, FAAD_MIN_STREAMSIZE*MAX_CHANNELS-bread, handle);
        if (tmpbread <= 0)
            sample_buffer = NULL; /* to make sure it stops now */
        bread+=tmpbread;
    } while (sample_buffer != NULL && !ctx->stop);

    NeAACDecClose(hDecoder);
#if defined STANDALONE and defined SDL
    SDL_CloseAudio();
#endif

    if (aac_buffer)
        free(aac_buffer);

    return frameInfo.error;
}

static int decodeMP2stream(decoder_context_t *ctx, URL_FILE *handle)
{
    unsigned char *sample_buffer=NULL;
    unsigned char *mp2_buffer=NULL;
    mpg123_handle *mh = NULL;
    int mc;
    size_t mp2_buffer_len=32768;
    size_t sample_buffer_len=256*1024;
    size_t bdecoded=0;
    int bread=0;
    int ret=-1;
    int err;
    long rate;
    int channels, format;

    if (!(mp2_buffer = (unsigned char*)malloc(mp2_buffer_len)))
    {
        fprintf(stderr, "Memory allocation error\n");
        goto errl;
    }
    if (!(sample_buffer = (unsigned char*)malloc(sample_buffer_len)))
    {
        fprintf(stderr, "Memory allocation error\n");
        goto errl;
    }
    
    if((mh = mpg123_new(NULL, &err)) == NULL || mpg123_open_feed(mh) != MPG123_OK) {
                fprintf(stderr, "Cannot initialize mpg123 library: %s", mpg123_plain_strerror(err));
		goto errl;
    } 
    do
    {
        bread = url_fread(mp2_buffer, 1, mp2_buffer_len, handle);
        if (bread <= 0) {
            ret=0;
            break;
        }
        mc=mpg123_decode(mh, mp2_buffer, bread, sample_buffer, sample_buffer_len, &bdecoded);
        if(mc == MPG123_NEW_FORMAT) {
            mpg123_getformat(mh, &rate, &channels, &format);
            fprintf(stderr, "Format MP2 %liHz %ich, format %i\n", rate, channels, format);
#if defined STANDALONE and defined SDL
	    desired.freq = rate;
    	    desired.format = AUDIO_S16;
    	    desired.channels = channels;
            desired.samples = 1024;        /* A good value for games */
            desired.callback = sdl_callback;
            desired.userdata = NULL;
            if ( SDL_OpenAudio(&desired, &obtained) < 0 ){
                fprintf(stderr, "Couldn't open audio: %s\n", SDL_GetError());
                exit(-1);
            }
            SDL_PauseAudio(0);
#endif
        }
        stream_cb(ctx->p, sample_buffer, bdecoded, rate, channels);
        
        while(mc != MPG123_ERR && mc != MPG123_NEED_MORE) { /* Get all decoded audio that is available now before feeding more input. */
            mc = mpg123_decode(mh, NULL, 0, sample_buffer, sample_buffer_len, &bdecoded);
            stream_cb(ctx->p, sample_buffer, bdecoded, rate, channels);
        }
        if(mc == MPG123_ERR){ 
            fprintf(stderr, "some error: %s", mpg123_strerror(mh)); 
            break; 
        }
    } while(!ctx->stop);

    mpg123_close(mh);
    mpg123_delete(mh);
#if defined STANDALONE and defined SDL
    SDL_CloseAudio();
#endif
errl:
    if(mp2_buffer)
        free(mp2_buffer);
    if(sample_buffer)
        free(sample_buffer);
    return ret;
}

void *decode_stream(void *p)
{
    decoder_context_t *ctx=(decoder_context_t *)p;
    char *url=ctx->url;

    if(!init_done) {
        int err;
        if((err=mpg123_init()) != MPG123_OK) {
            fprintf(stderr, "Cannot initialize mpg123 library: %s", mpg123_plain_strerror(err));
            return NULL;
        }
        init_done=1;
    }
    do {
        URL_FILE *handle;
        fprintf(stderr, "---> URL open: %s\n",url);
        handle = url_fopen(url, "r");
        unsigned char buffer[2];
        int bread;
    
        if (handle == NULL) {
            /* unable to open file */
            fprintf(stderr, "Error opening url: %s\n", url);
#if defined STANDALONE
            return NULL;
#else
            continue;
#endif                        
        }
        fprintf(stderr, "---> OK url: %s\n", url);
        
        bread = url_fread(buffer, 1, sizeof(buffer), handle);
    
        if(bread==sizeof(buffer)) {
            url_rewind(handle);
            if(!memcmp(buffer, "\xff\xfc",2)) {
                decodeMP2stream(ctx, handle);
            } else if(!memcmp(buffer, "\xff\xf9",2)) {
                decodeAACstream(ctx, handle);
            } else {
                fprintf(stderr, "No valid stream at %s found!\n", url);
            }
        }
        if(handle) {
            fprintf(stderr, "---> closed url: %s\n", url);
            url_fclose(handle);
        }
        usleep(1000000);
    } while(!ctx->stop);
    return NULL;
}


#ifdef STANDALONE

int main(int argc, char *argv[]) {
    decoder_context_t d;
#if defined STANDALONE and defined SDL
    SDL_Init(SDL_INIT_AUDIO);
    ctx=bs1770_ctx_open_default(1);
#endif
    pcm=(char *)malloc(BUFSZ);
    register_stream_callback(write_cb);
    d.url=argv[1];
    d.stop=0;
    decode_stream(&d);
    mpg123_exit();
    free(pcm);
#if defined STANDALONE and defined SDL
    SDL_Quit();
    bs1770_ctx_close(ctx);
#endif
    return 0;
}
#endif

